/* 
GA for Skill based Resource Allocation

  Author : Dipti Gagneja (di_gagneja@hotmail.com)
  Done as part of AI Assignment under the guidance of Dr. S.B. Nair , Year : 2006

File Name :ga.c 

Abstract
 Defines the functions implementing the individual steps for the GA - Initializing the population, Calculating fitness, 
 Crossover, Mutation, Selecting New generation and other utility functions 
*/

#include <math.h>
#include "ga.h"
#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>


/********************************INITAILIZE A POPULATION**************/

/*FUNCTIONS-  Initialize, create_gene ,validate */

void initialize()
{
	int counter=1;
	vflag=malloc(no_of_employees*4);
	p=malloc((no_of_skills+1)*8);
	do
	{
		start = (header*) malloc(sizeof(header));
		/*Initially the fitness of each chromosome is set to -1*/
		start->fitness=-1;
		start->first=NULL;
		temp = (gene*) malloc(sizeof(gene));
		start->first = temp;
		create_gene(temp,1,0,0);
		population[counter]=start;
		validate(population[counter]);
		counter++;
		
	}while(counter<=population_size);
	
}

/*Function create_gene
Recursive function, generates the genes in a chromosome. 
Each gene is made of randomly generated employee id and a project id 
Arguments  temp					- pointer to the current gene, 
		   current_project_id   - project id in the current gene
		   no_of_genes			- no of genes generated so far
		   team_size			- team size of the project with id = current_project_id */

void create_gene(gene *temp,int current_project_id,int no_of_genes,int team_size)
{
	no_of_genes++;
	team_size++;
	temp->project_id = current_project_id;
	temp->employee_id=ga_random_int(1,no_of_employees);
/*if the no of geen generated reaches the total req
 i.e. the length of chromosome, set the next pointer as null */
	if(no_of_genes==total_req) 
		temp->next=NULL;
	else
	{
		temp->next=(gene*) malloc(sizeof(gene));
		if(team_size==project_team_size[current_project_id])
		/*genes for a particular project have been made, 
		call the function for next project*/
			create_gene(temp->next,current_project_id+1,no_of_genes,0);
		else
			create_gene(temp->next,current_project_id,no_of_genes,team_size);
	}
	return;
}

/*function to validate 
that one employee is allocated to at most one project*/

void validate(header *start)
{
	int counter,random_int;

	for(counter=1;counter<=no_of_employees;counter++)
		vflag[counter]=0;

	tempd = start->first;
	while(tempd!=NULL)
	{

		if(vflag[tempd->employee_id]==1)
		{
		/*if the employee is already allocated to an earlier project 
		then re-generate a random employee id not generated till now 
		and set the employee id to that*/
			random_int=ga_random_int(1,no_of_employees);

			while(vflag[random_int]!=0)
				random_int=ga_random_int(1,no_of_employees);
			
			tempd->employee_id = random_int;
			vflag[tempd->employee_id]=1;
		}
		else
			vflag[tempd->employee_id]=1;
		
		tempd=tempd->next;
	}
	

	
}


/****************************FITNESS CALCULATION***********************/

/*Function - cal_fitness to calculate the fitness of a chromosome (fitness = 1/ (Total Penalty))
calls functions which return the penalties due to violation of different constraints*/

double cal_fitness(header *start)
{
	
	double skill_p,dist_p,budget_p,pref_p;
	skill_p=skill_match(start);
	dist_p=distribution_penalty(start);
	budget_p=budget_penalty(start);
	pref_p = preference_penalty(start);
	start->fitness = wt_skill*skill_p + dist_p +wt_budget* budget_p + wt_pref*pref_p;
	if(dist_p>0)
/*If distribution is not valid (penalty greater than '0') then assign a very high penalty 
(so as the chromosome is lost during evolution*/
			start->fitness *=100;
	start->fitness = 1/start->fitness;
	return (start->fitness);
}


/*Function skill_match 
Returns penalty due skill mismatch among employee skills and project requirements 
Argument - header node of the chromosome */

double skill_match(header *start)
{
	int counter;
	double ret_val=0.00;
		
	for(counter=1;counter<=no_of_skills;counter++)
		p[counter]=0.0;

	tempf = start->first;
	
	while(tempf!=NULL)
	{
		for(counter=1;counter<=no_of_skills;counter++)
		{
/*penalty for SKILL 'counter' is difference of level req by project and that of the employee*/
			p[counter]=(double)(project_skill[counter][tempf->project_id] - employee_skill[counter][tempf->employee_id]);

/*If skill is absent in the employee allocated then increase the penalty by 50%*/
			if(employee_skill[counter][tempf->employee_id] ==0 && project_skill[counter][tempf->employee_id] !=0) //skill absent 
				p[counter]=abs(p[counter])*1.5;

/*If the employee is overqualified then decrease the penalty by 50%*/
			if(p[counter]<0)
				p[counter] = abs(p[counter])*0.5;
			
			ret_val+=p[counter];
		}
		
		tempf=tempf->next;
	}

	
	return ret_val;
}

/*Function Budget_penalty
Returns penalty ( = percentage of overflow )due to budget overflow
Argument - header node of the chromosome */

double budget_penalty(header *start)
{
	int counter;
	double proj_cost,temp_cost;
	double ret_val=0.00;
	for(counter=1;counter<=no_of_projects;counter++)
	{
		
		proj_cost = project_cost1(counter,start);
		temp_cost=project_budget[counter]-proj_cost;
		if(temp_cost<0)
			ret_val+=(((proj_cost-project_budget[counter])/project_budget[counter])*100);
	}
	return ret_val;
}

/*Function project_cost1 
Returns the cost of a project
Arguments - project id and the header pointer of the chromosome under consideration*/

double project_cost1(int project_id ,header *start)
{
	double ret_val=0.00;

	tempf = start->first;
	while(tempf!=NULL)
	{
		
		if(tempf->project_id ==project_id)
		{
			ret_val+=(employee_cost[tempf->employee_id]*project_hours[project_id]);
		}
		tempf=tempf->next;
	}
	return ret_val;
}


/*Function Distributed_penalty 
Returns penalty due to wrong distribution of employees in a chromosome 
Arguments - header of the chromosome under consideration*/

double distribution_penalty(header *start)
{
	int *flag,counter,ret_val=0;
	float temp_penalty;
	flag = malloc(no_of_employees*4);
	tempf=start->first;

	for(counter=1;counter<=no_of_employees;counter++)
		flag[counter]=0;
	
	while(tempf!=NULL)
	{
		if(flag[tempf->employee_id]==1)
			ret_val++;
		else
			flag[tempf->employee_id]=1;
					
		tempf=tempf->next;
	}
	
/*Penalty= percentage of employees allocated to more than one project wrt total requirement*/
	temp_penalty= ((float)ret_val/(float)total_req);
	return (temp_penalty*100);
}


/*Function preference_penalty 
Returns the penalty due to allocation of non preferred projects to employees
Argument - the header pointer of the chromosome under consideration*/

double preference_penalty(header *start)
{
	double ret_val=0.00;
	tempf=start->first;
	while(tempf!=NULL)
	{
		if(pref_project[tempf->employee_id]!=tempf->project_id)
			ret_val++;
		tempf=tempf->next;
	}

/*Penalty = Percentage of employees allocated non preferred project*/

	ret_val=(ret_val/total_req)*100;

	return ret_val;
}



/*****Sorting of the chromosomes in the population according to decreasing fitness******/	

/*Function Arrange 
arguments - Population array and the number of chromosomes*/

void arrange(header *pop[],int no_of_chr)
{
	int counter1,counter2;
	tempa=NULL;
	for(counter1=1;counter1<=no_of_chr-1;counter1++)
	{
		for(counter2=counter1+1;counter2<=no_of_chr;counter2++)
		{
			if(pop[counter2]->fitness>pop[counter1]->fitness)
			{
				tempa=pop[counter1];
				pop[counter1]=pop[counter2];
				pop[counter2]=tempa;
			}
		}
	}
}




/*void print_chromosome()
{
	int i;
	for(i=1;i<=population_size;i++)
	{
		printf("\n\nChromosome : %d\t Fitness  : %0.2f\n",i,population[i]->fitness);
		tempp=population[i]->first;
		while(tempp!=NULL)
		{
			printf(" %d-%d ",tempp->project_id,tempp->employee_id);
			tempp=tempp->next;
		}
	}
}*/

/*************************GENETIC OPERATORS *****************************/

/*******************CROSSOVER **********************/
/*FUNCTIONS  - crossover, set_parent, rotate*/

void crossover()
{
	float prob;
	int counter1=0,counter2=0,swap=0,crossover_point=0,no_of_replace=0,child_array_counter=0;
	int rotate_amount=0;

	no_of_children=0;

	for(counter1=1;counter1<=population_size;counter1++)
		parent[counter1]=population[counter1];

	counter1=1;
	while(counter1<population_size)
	{
			prob=ga_random_float();
		
			if(prob<=pC)
			{
				
/*Set the parents for crossover - ith and i+1th chromosome in the ordered list of chromosome*/

				set_parent(population[counter1],1);
				set_parent(population[counter1+1],2);
			
/*Rotate the second parent by random amount*/

				rotate_amount = ga_random_int(1,total_req-1);
				rotate_c(parent2,rotate_amount);
				tempp=parent2->first;

/*Randomly select a crossover point*/
				
				crossover_point= ga_random_int(1,total_req-1);

/*Create two child chromosomes*/

				child1=malloc(sizeof(header));
				child2=malloc(sizeof(header));
				tempr=malloc(sizeof(gene));
				tempp=malloc(sizeof(gene));
				child1->first=tempr;
				child2->first=tempp;
				child1->fitness=-1;
				child2->fitness=-1;

				temp1=parent1->first;
				temp2=parent2->first;
						
				while(counter2<crossover_point)
				{
/*Copy the parent chromosomes till the crossover point to the child chromosomes*/

					counter2++;
					tempr->employee_id= temp1->employee_id;
					tempr->project_id = temp1->project_id;
					tempr->next = malloc(sizeof(gene));
					tempr=tempr->next;
				
					tempp->employee_id= temp2->employee_id;
					tempp->project_id = temp2->project_id;
					tempp->next = malloc(sizeof(gene));
					tempp=tempp->next;

					temp1=temp1->next;
					temp2=temp2->next;
				}

							
				while(temp1->next!=NULL)
				{
/*Interchange the genes of the parent chromosomes after the crossover point*/

					tempr->employee_id= temp2->employee_id;
					tempr->project_id = temp1->project_id;
					tempr->next = malloc(sizeof(gene));
					tempr=tempr->next;
				
					tempp->employee_id= temp1->employee_id;
					tempp->project_id = temp2->project_id;
					tempp->next = malloc(sizeof(gene));
					tempp=tempp->next;

					temp1=temp1->next;
					temp2=temp2->next;
				}

				tempr->employee_id= temp2->employee_id;
				tempr->project_id = temp1->project_id;
				tempr->next = NULL;


				tempp->employee_id= temp1->employee_id;
				tempp->project_id = temp2->project_id;
				tempp->next = NULL;

/*Rotate back the second child so that the configuration of both children is the same */
				rotate_c(child2,total_req-rotate_amount);

/*Calculate the fitness of each of the child*/
				child1->fitness=cal_fitness(child1);
				child2->fitness=cal_fitness(child2);

/*Add the children just generated to the children array*/
				child_array_counter++;
				children[child_array_counter]=child1;
				child_array_counter++;
				children[child_array_counter]=child2;
				
			}

			counter1=counter1+2;
	}
	no_of_children = child_array_counter;
}

/*Function Set_parent -  Sets the chromosomes as parents 
Arguments - header point of the chromosome, which parent to be set - as parent 1 or 2 */

void set_parent(header *start,int set_id)
{
	if(set_id==1)
	{
			parent1=malloc(sizeof(header));
			tempp=malloc(sizeof(gene));
			parent1->first=tempp;
			parent1->fitness = start->fitness;
			tempr = start->first;
			while(tempr->next!=NULL)
			{
				tempp->employee_id= tempr->employee_id;
				tempp->project_id = tempr->project_id;
				tempp->next = malloc(sizeof(gene));
				tempp=tempp->next;
				tempr=tempr->next;
			}
			tempp->employee_id= tempr->employee_id;
			tempp->project_id = tempr->project_id;
			tempp->next=NULL;
	}
	if(set_id==2)
	{
			parent2=malloc(sizeof(header));
			tempp=malloc(sizeof(gene));
			parent2->first=tempp;
			parent2->fitness = start->fitness;
			tempr = start->first;
			while(tempr->next!=NULL)
			{
				tempp->employee_id= tempr->employee_id;
				tempp->project_id = tempr->project_id;
				tempp->next = malloc(sizeof(gene));
				tempp=tempp->next;
				tempr=tempr->next;
			}
			tempp->employee_id= tempr->employee_id;
			tempp->project_id = tempr->project_id;
			tempp->next=NULL;
	}
}

/*Function rotate_c - Rotate a chromosome
Arguments - pointer to header node of the chromosome and 
		   the amount by which the chromosome is to be rotated*/

void rotate_c(header *start,int amount_to_rotate)
{
	int number_of_genes=0, temp_counter=0;
	
	
	tempv=start->first;
	tempr=start->first;
	while(tempv->next!=NULL)
	{
		number_of_genes++;
		tempv=tempv->next;
	}
	while(temp_counter!=number_of_genes-amount_to_rotate)
	{
		temp_counter++;
		tempr=tempr->next;
	}
	
	
	tempv->next= start->first;
	start->first=tempr->next;
	tempr->next=NULL;
}


/******************MUTATION**********************/

/*FUNCTION - mutation*/

void mutation()
{
	int counter=1,prob1,prob2,swap=0 ;
	float mut_prob;
	tempd=NULL;
	tempr=NULL;
	for(counter=1;counter<=no_of_children;counter++)
	{
		mut_prob = ga_random_float();
		if(mut_prob<=pM)
		{
				tempd=children[counter]->first;
				tempr=children[counter]->first;
				prob1=ga_random_int(1,no_of_projects);
				prob2=ga_random_int(1,no_of_projects);
				while(prob2==prob1)
					prob2=ga_random_int(1,no_of_projects);
				
				while(tempd->project_id!=prob1)
						tempd=tempd->next;
				while(tempr->project_id!=prob2)
						tempr=tempr->next;

/*Swap one of the employees allocated to any two projects*/

				swap = tempd->employee_id;
				tempd->employee_id = tempr->employee_id;
				tempr->employee_id =swap;
				children[counter]->fitness= cal_fitness(children[counter]);
				
		}
		
	}
}
		

/*******************SELECTION OF NEW GENERATION*******************/

/*Select best chromosomes (=population size) from the parent and children arrays */

void new_generation()
{
	int pop_counter,child_counter=1,parent_counter=1;
	for(pop_counter=1;pop_counter<=population_size;pop_counter++)
	{
		if(child_counter<=no_of_children)
		{
			if(parent[parent_counter]->fitness>children[child_counter]->fitness)
			{
				population[pop_counter]=parent[parent_counter];
				parent_counter++;
			}
			else
			{
				population[pop_counter]=children[child_counter];
				child_counter++;
			}
		}
		else
		{
				population[pop_counter]=parent[parent_counter];
				parent_counter++;
		}
	}
}


/**********************Calculate the average penalty of a population**********/

double penalty_stat()
{
	int counter=1;
	double penalty_sum=0.00,penalty_avg=0.00;
	temp = NULL;
	for(counter=1;counter<=population_size;counter++)
	{
		penalty_sum=penalty_sum + population[counter]->fitness;
	}

	penalty_avg = penalty_sum/population_size;

	return penalty_avg;
}



/*************RANDOM number generation *************/

int ga_random_int(int start, int end)
{
	init1(rand());
	return (rand1()%end+start);
}

float ga_random_float()
{
	init1(rand());
	return ((float)rand1() / 4294967296.0);
}

/********************ALTER THE TEAM SIZE AND BUDGET *************/
/*The Function is invoked only in case the number of employees available are 
less as compared to the requirement*/
/*It decreases the immediate required number of employees for each project, 
the higher profit projects are given higher number of employees */
void alter_team_size()
{
	int counter, new_total_req,max,temp_team_size;
	new_team_size = malloc(no_of_projects*4);
	flagt = malloc(no_of_projects*4);
	for(counter=1;counter<=no_of_projects;counter++)
	{
		new_team_size[counter]=project_team_size[counter];
		flagt[counter] = 0;
	}
	
	do
	{  /*Decrease the team size of each project till the requirement
	   is more than the availability*/
		for(counter=1;counter<=no_of_projects;counter++)
		{
			if(new_team_size[counter] > 1)
				new_team_size[counter]--;
		}
		new_total_req =0;
		for(counter=1;counter<=no_of_projects;counter++)
				new_total_req += new_team_size[counter];
			
	}while(new_total_req > no_of_employees);

	while(new_total_req - no_of_employees <0)
	{
		/*Allocated the extra employees is any to the higher profit projects*/
		max =1;
		for(counter=2;counter<=no_of_projects;counter++)
		{
			if(project_profit[max] < project_profit[counter] && flagt[counter]==0)
				max = counter;
		}
	
		flagt[max]=1;
		temp_team_size= project_team_size[max]-new_team_size[max];
		if(temp_team_size<= new_total_req - no_of_employees)
			new_team_size[max] = project_team_size[max];
		else
			new_team_size[max] +=(no_of_employees - new_total_req);
		
		new_total_req = 0;
		for(counter=1;counter<=no_of_projects;counter++)
				new_total_req+=new_team_size[counter];
		
	}
	printf("The present requirements of the projects\n\n");
	printf("Project Id\tteam size\tBudget\t\tProfit\n");

	for(counter=1;counter<=no_of_projects;counter++)
	{
		printf("%d\t\t%d\t\t%0.2f",counter,project_team_size[counter],project_budget[counter]);
		if(project_budget[counter] <10000.00)
			printf("\t\t%4.2f\n",project_profit[counter]);
		else
			printf("\t%4.2f\n",project_profit[counter]);
	}

	printf("\nChanging the team size and the budget, continuing wih the following data\n\n");
	printf("Project Id\tteam size\tBudget\t\tProfit\n");
	
	for(counter=1;counter<=no_of_projects;counter++)
	{
		/*Change the budget availability proportional to the decreased team sze*/

		project_budget[counter] = (project_budget[counter] / project_team_size[counter]) * new_team_size[counter];
		project_team_size[counter] = new_team_size[counter];
		printf("%d\t\t%d\t\t%0.2f",counter,project_team_size[counter],project_budget[counter]);
		if(project_budget[counter] <10000.00)
			printf("\t\t%4.2f\n",project_profit[counter]);
		else
			printf("\t%4.2f\n",project_profit[counter]);
	}
	total_req = new_total_req;
}


